#include <fstream>
#include <sstream>
#include <stdlib.h>

#include "xma_parse.h"
#include "Bit_stream.h"

using namespace std;

namespace Get_arguments {

struct Argument_error {
    string description;
    Argument_error(string desc) : description(desc) {}
};

void get(int argc, char * argv[], const char ** filename, const char ** out_filename, long *offset, long *block_size, long *data_size, int *channels, bool *verbose);
}

int main(int argc, char * argv[]) {
    const char * input_filename, * output_filename;
    long offset, block_size, data_size;
    int channels;
    bool verbose;

    cout << "Experimental XMA parser 0.3 (blame hcs)" << endl << endl << flush;

    try {
        Get_arguments::get(argc, argv, &input_filename, &output_filename, &offset, &block_size, &data_size, &channels, &verbose);
    } catch (Get_arguments::Argument_error ae) {
        cerr << ae.description << endl
    << "usage: xma_parse filename [-o offset] [-b block size] [-d data size]\n"
    << "       [-x output filename] [-v]" << endl;
        exit(EXIT_FAILURE);
    }

    ifstream is(input_filename, ios::binary);

    if (!is) {
        cerr << "error opening file!" << endl;
        exit(EXIT_FAILURE);
    }

    if (-1 == data_size) {
        is.seekg(0, ios::end);
        data_size = static_cast<long>(is.tellg()) - offset;
    }

    ofstream os;

    if (NULL != output_filename)
    {
        os.open(output_filename, ios::binary);

        if (!os) {
            cerr << "error opening output file!" << endl;
            exit(EXIT_FAILURE);
        }
    }

    cout << "filename: " << input_filename << endl;
    if (os.is_open()) {
        cout << "output filename: " << output_filename << endl;
    }
    cout << "offset: " << hex << offset << dec << endl;
    cout << "block size: " << hex << block_size << dec << endl;
    cout << "data size: " << hex << data_size << dec << endl;
    cout << "------------------" << endl << endl;

    try {
        unsigned long total_sample_count = 0;
        for (long block_offset = offset;
                  block_offset < offset + data_size;
                  block_offset += block_size ) {

            long usable_block_size = block_size;

            if (block_offset + usable_block_size > offset + data_size)
            {
                usable_block_size = offset + data_size - block_offset;
            }

            unsigned long sample_count = Parse_XMA::parse_block(is, os, block_offset, usable_block_size, channels, verbose);
            if (verbose) {
            cout << endl << sample_count << " samples (block)" << endl << endl;
            }

            total_sample_count += sample_count;
        }
        cout << endl << total_sample_count << " samples (total)" << endl;
    }
    catch (const Bit_stream::Out_of_bits& oob) {
        cerr << "error reading bitstream" << endl;
        exit(EXIT_FAILURE);
    }
    catch (const Parse_XMA::Parse_error& pe) {
        cerr << pe << endl;
        exit(EXIT_FAILURE);
    }
}

void Get_arguments::get(int argc, char * argv[], const char ** filename, const char ** out_filename, long *offset, long *block_size, long *data_size, int *channels, bool *verbose) {
    if (argc < 2) {
        throw Argument_error("missing file name");
    }
    *filename = argv[1];
    *out_filename = NULL;
    *offset = 0;
    *block_size = 0x8000;
    *channels = 1;
    *data_size = -1;
    *verbose = false;

    for (int argno = 2; argno < argc; argno++) {
        if (argv[argno][0] == '-' && argv[argno][1] != '\0' &&
                argv[argno][2] == '\0') {
            switch (argv[argno][1]) {
                case 'o':
                    if (argno + 1 >= argc) {
                        throw Argument_error("-o needs an argument");
                    }

                    {
                        stringstream argstring(argv[++argno]);
                        argstring >> hex >> *offset;
                        /* TODO: check that conversion went ok */
                    }
                    continue;
                case 'b':
                    if (argno + 1 >= argc) {
                        throw Argument_error("-b needs an argument");
                    }

                    {
                        stringstream argstring(argv[++argno]);
                        argstring >> hex >> *block_size;
                        /* TODO: check that conversion went ok */
                    }
                    continue;
                case 'c':
                    if (argno + 1 >= argc) {
                        throw Argument_error("-c needs an argument");
                    }
                    {
                        stringstream argstring(argv[++argno]);
                        argstring >> *channels;
                        /* TODO: check that conversion went ok */
                    }
                    continue;
                case 'd':
                    if (argno + 1 >= argc) {
                        throw Argument_error("-d needs an argument");
                    }
                    {
                        stringstream argstring(argv[++argno]);
                        argstring >> hex >> *data_size;
                        /* TODO: check that conversion went ok */
                    }
                    continue;
                case 'x':
                    if (argno + 1 >= argc) {
                        throw Argument_error("-x needs an argument");
                    }
                    *out_filename = argv[++argno];
                    continue;
                case 'v':
                    *verbose = true;
                    continue;
                default:
                    throw Argument_error(
                            string("don't know what to do with argument \"") + argv[argno] + "\"");
            }
        }

        throw Argument_error(
                string("don't know what to do with argument \"") + argv[argno] + "\"");
    }
}

